/*
 * powwow.h
 * various constants etc
 */

#ifndef LONG_MIN
#define LONG_MIN ((long)((unsigned long)1L << sizeof(long) * 8 - 1))
#endif

#ifndef LONG_MAX
#define LONG_MAX ((long)(((unsigned long)1L << sizeof(long) * 8 - 1) - 1))
#endif

#define uSEC_PER_SEC 1000000L  /* microseconds in a second */
#define mSEC_PER_SEC 1000L     /* milliseconds in a second */
#define uSEC_PER_mSEC 1000L    /* microseconds in a millisecond */

#ifdef MCH_AMIGA
#define EINTR 4
#include <exec/types.h>
#include <pragmas/socket_lib.h>
#endif

#ifndef _H_SYS_TIME
#include <sys/time.h>
#endif

#ifndef __STDIO_H
#include <stdio.h>
#endif

typedef struct timeval vtime;   /* needs #include <sys/tyme.h> */

typedef struct 
{               		/* for additional connected sessions */
  int active;			/* display remote output? */
  char *id;			/* session id */
  char *host;                   /* address of remote host */
  int port;                     /* port number of remote host */
} consess;

/* generic linked list node (never actually created) */
typedef struct defnode {
    struct defnode *next;
    char *sortfield;
} defnode;

/* linked list nodes: keep the next pointer first, then string to sort by */
typedef struct aliasnode {
    struct aliasnode *next;
    char *name;
    char *subst;
} aliasnode;

typedef struct actionnode {
    struct actionnode *next;
    char *pattern;
    char *command, *label;
    int active;
} actionnode;

typedef struct marknode {
    struct marknode *next;
    char *pattern;
    int attrcode;
} marknode;

typedef struct keynode {
    struct keynode *next;	/* next key definition in list */
    char *name;			/* key name */
    char *sequence;		/* escape sequence sent by terminal */
#ifdef MCH_AMIGA
    UWORD code,qualifier; /* raw key data; -1 -1 for non-Amiga sequences */
#define KEYBIND_MAGIC 0xffff
#endif
    void (*funct)();		/* function called when key pressed */
    char *call_data;		/* data passed to function */
} keynode;

typedef struct delaynode {
    struct delaynode *next;
    char *name;
    char *command;
    vtime when;                 /* structure containing time when */
				/* command must be executed */
} delaynode;

typedef struct varnode {        /* for named variables */
    struct varnode *next;
    char *name;
    int index;
    long num;
    char *str;
} varnode;

/* shared functions: */

/* from powwow.c */
int read_settings();
int save_settings();
void setsigs();
void connect_sess();
void disconnect_sess();
void show_connections();
void show_host();
void exitpowwow();
void execute();
void syserr();
char *skipspace();
void process_remote_input();
void put_marks();
void wrap_print();
void smart_print();
void suspend_powwow();
int is_main_sess();
int main_sess();
int find_sess();
void send_to_main_host();
void setfile();
void strdel();
void strins();
void showprompt();
void parse_user_input();
char *parse_instruction();
void to_input_line();
void cancel_edit();
void set_param();
void restore_param();
void subst_param();
void printver();
void snoop_sess();
void flashparen();
void put_in_history();
void put_word();
char *split_first_word();
int printstrlen();
int parse_walk();

/* from xeval.c */
void my_memmove();

/* from mapping.c */
void retrace_steps();
void show_automap();
void print_automap();
void add_dir_to_map();

/* from tel.c */
int doconnect();
int read_socket();
void send_to_host();
void flush_output_buffer();
void send_term_size();

/* from list.c */
long diff_vtime();
void add_vtime();
int cmp_vtime();
int time_sort();
int rev_time_sort();
void add_node();
void add_aliasnode();
void add_actionnode();
void add_marknode();
void add_keynode();
void add_delaynode();
varnode *add_varnode();
aliasnode **lookup_alias();
actionnode **lookup_action();
actionnode **lookup_action_pattern();
marknode **lookup_marker();
keynode **lookup_key();
delaynode **lookup_delay();
varnode **lookup_varnode();
void delete_aliasnode();
void delete_actionnode();
void delete_marknode();
void delete_keynode();
void delete_delaynode();
void delete_varnode();

/* from beam.c */
void message_edit();
int process_message();
void childdeathsig();

#define PRINT_INT if(!internal) internal=1; printf

#define MAXFDSCAN 32		/* max number of fd to scan in select() */
#define MAXHIST 128		/* number of history lines kept */
#define NIL_CHAR '`'            /* special escape char (there is also \ ) */
#define DELIM " ;"		/* word delimiting characters */
#define SPECIAL_CHARS "{}();\"=" /* specials chars needing escape */
#define CAPLEN 20		/* max length of a terminal capability */
#define BUFSIZE 4096		/* general buffer size */
#define OUTBUFSIZE 4096         /* output buffer size */
#define LINELEN 1024		/* maximum line length */
#define PARAMLEN 1024		/* maximum length of strings (&x, $x) */
#define NUMPARAM 10             /* number of different params allowed */
#define NUMVAR 50               /* number of global variables allowed */
#define MAXMAPLEN 500		/* maximum length of automapped path */
#define WORDSKEPT 512		/* number of words kept for TAB-completion */
#define MPI "~$#E"		/* MUME protocol introducer */
#define SOCKBUFSIZE 4096	/* socket buffer size for read */
#define MAXSUBOPT 256		/* max length of suboption string */
#define MAXARGS 16		/* max number of arguments to editor */
#define FLASHDELAY 500	        /* time of parentheses flash in millisecs */
#define CMDSEP ';'		/* command separator character */
#define KBD_TIMEOUT 35000	/* timeout for keyboard read in microsecs;
				   35000 should be enough for 300+ baud */

#define MAX_STACK   100         /* maximum number of nested
                                 * action, alias, #for or #while */
#define MAX_LOOP  10000         /* maximum number of iterations in
                                 * for or #while */

/* Variable struct definitions */
typedef struct {                       /* holds value of local parameters */
  long num[NUMPARAM];
  char str[NUMPARAM][PARAMLEN];
} params;

typedef struct {                       /* vector element of pointers */
  long *num;                           /* to ALL variables */
  char *str;
} vars;

typedef struct {
  params *p[MAX_STACK];
  int curr;
} param_stack;

#define LM_NOECHO 1	/* no local echo */
#define LM_CHAR 2	/* char-by-char mode (no line editing) */

/* editing session control */
typedef struct editsess_ 
{
  int key;		/* session identifier (-1 if cancelled) */
  char builtin;         /* 1 if viewing built-in text */
  int pid;		/* pid of child */
  char *descr;  	/* short description of what we are editing */
  char *file;		/* name of temporary file */
  time_t ctime; 	/* time when temp file was created (upper bound) */
  long oldsize; 	/* original file size */
  struct editsess_ *next;
} editsess;

/* shared vars */
/* from powwow.c */
extern int internal;
extern char history_done;
extern int linemode;		/* line mode with flags as above (LM_*) */
extern int sockfd;		/* socket file descriptor */
extern char hostname[], portnumber[];
extern char deffile[], helpfile[];

extern int cols, lines;		/* terminal window size */
extern int line0, col0;		/* origin of input line */
extern char promptstr[];	/* prompt string */
extern int promptlen;           /* length of prompt string */
extern char promptintercept;    /* 1 if prompt was intercepted by an #action */
extern char edbuf[];		/* input line buffer */
extern int edlen;		/* characters in edbuf */
extern int pos;			/* cursor position in edbuf */
extern char edattrbeg[], edattrend[];

extern long received, sent;

#ifndef MCH_AMIGA
extern clock_t start_clock, cpu_clock;
#endif

extern aliasnode *aliases;
extern actionnode *actions;
extern marknode *markers;
extern int a_nice;
extern keynode *keydefs;
extern delaynode *delays;
extern delaynode *dead_delays;
extern varnode *named_vars[2];
extern int num_named_vars[2];
extern int max_named_vars;
extern vars *var;
#ifndef MCH_AMIGA
extern vtime now, start_time, ref_time;
#endif

extern char initstr[];
extern FILE *capfile, *recordfile;
extern char prefixstr[];
extern char inserted_next[];
extern char verbatim;

extern char speedwalk;
extern char compactmode;
extern char wrap;
extern char echo_int;
extern char echo_ext;
extern char debug;
extern int also_history_words;

extern char *hist[];
extern int curline;
extern int pickline;

/* from beam.c */
extern char edit_start[];
extern char edit_end[];
extern editsess *edit_sess;

#define CONTROL(c) ((c) & 0x1f)

/* macros to find cursor position from input buffer position */
#define CURLINE(pos) (((pos) + col0) / (cols - 1) + line0)
#define CURCOL(pos) (((pos) + col0) % (cols - 1))

/* macros to match parentheses */
#define ISRPAREN(c) ((c) == ')' || (c) == ']' || (c) == '}')
#define ISLPAREN(c) ((c) == '(' || (c) == '[' || (c) == '{')
#define LPAREN(c) ((c) == ')' ? '(' : ((c) == ']' ? '[' : '{'))

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

/*
 * Attribute codes: bit 0-4 for foreground color or effects, 5-9 for
 * background color. Color #16 is "none", so 0x0210 is "no attribute".
 */
#define COLORS         16         /* number of colors on HFT terminal */
#define LOWCOLORS      8          /* number of ANSI colors */
#define NO_COLOR       COLORS     /* no color change, use default */
#define BITS_COLOR     5          /* bits used for a color entry
                                     (16==none is a valid color) */
#define BITS_2COLOR    10         /* bits used for 2 color entries */
#define COLOR_MASK     0x1F       /* 5 (BITS_COLOR) bits set to 1, others 0 */

#define ATTR_BOLD      0x01
#define ATTR_BLINK     0x02
#define ATTR_UNDERLINE 0x04
#define ATTR_INVERSE   0x08

/*
 * WARNING: colors and attributes are currently using 14 bits
 * out of 16 (type used is int) and 0xFFFFFFFF ==  -1 is used
 * as 'invalid attribute' so there is only 1 bit left unused.
 */

/* constructors / accessors for attribute codes */
#define ATTRCODE(attr, fg, bg) \
                (((attr) << BITS_2COLOR) | ((bg) << BITS_COLOR) | (fg))
#define FOREGROUND(attrcode) ((attrcode) & COLOR_MASK)
#define BACKGROUND(attrcode) (((attrcode) >> BITS_COLOR) & COLOR_MASK)
#define ATTR(attrcode)       ((attrcode) >> BITS_2COLOR)

#define NOATTRCODE ATTRCODE(0, NO_COLOR, NO_COLOR)

/*
 * NCSA telnet 2.2 doesn't reset the color when it receives "esc [ m",
 * so we must know what the normal colors are in order to reset it.
 * These colors can be changed with the #color command.
 */
#ifdef TELNETBUG
#define DEFAULTFG 7		/* make default white text */
#define DEFAULTBG 4		/* on blue background */
#endif
